@@ -27,7 +27,7 @@
 #include <net/if.h>
 #endif
 
-#ifndef WIN32
+#if !defined(WIN32) && !defined(PS2_EE)
 #include <sys/socket.h>  /* struct sockaddr_storage */
 #endif
 
@@ -35,6 +35,15 @@
 #define IFNAMSIZ 255
 #endif
 
+#if defined(PS3_PPU) && !defined(IFNAMSIZ)
+#define IFNAMSIZ 16
+#endif
+
+#if defined(PS2_EE) && !defined(IFNAMSIZ)
+#define IFNAMSIZ 16
+#endif
+
+#include "libnfs-multithreading.h"
 #include "libnfs-zdr.h"
 #include "../nfs/libnfs-raw-nfs.h"
 #include "../nfs4/libnfs-raw-nfs4.h"
@@ -50,7 +59,7 @@ extern "C" {
 #define MAX(a, b) (((a) > (b)) ? (a) : (b))
 #endif
 
-#if !defined(HAVE_SOCKADDR_STORAGE) && !defined(WIN32)
+#if !defined(HAVE_SOCKADDR_STORAGE) && !defined(WIN32) && !defined(PS2_EE)
 /*
  * RFC 2553: protocol-independent placeholder for socket addresses
  */
@@ -123,6 +132,9 @@ struct rpc_context {
 	struct sockaddr_storage udp_src;
 	struct rpc_queue waitpdu[HASHES];
 	uint32_t waitpdu_len;
+#ifdef HAVE_MULTITHREADING
+        libnfs_mutex_t rpc_mutex;
+#endif /* HAVE_MULTITHREADING */
 
 	uint32_t inpos;
 	char rm_buf[4];
@@ -205,12 +217,16 @@ void nfs_set_error(struct nfs_context *nfs, char *error_string, ...)
 #endif
 ;
 
+#if defined(PS2_EE)        
+#define RPC_LOG(rpc, level, format, ...) ;
+#else        
 #define RPC_LOG(rpc, level, format, ...) \
 	do { \
 		if (level <= rpc->debug) { \
 			fprintf(stderr, "libnfs:%d " format "\n", level, ## __VA_ARGS__); \
 		} \
 	} while (0)
+#endif
 
 const char *nfs_get_server(struct nfs_context *nfs);
 const char *nfs_get_export(struct nfs_context *nfs);
@@ -271,35 +287,57 @@ struct nfs_fh {
         char *val;
 };
 
-struct nfs_context {
-       struct rpc_context *rpc;
+struct nfs_context_internal {
        char *server;
        char *export;
+       char *cwd;
        struct nfs_fh rootfh;
        uint64_t readmax;
        uint64_t writemax;
-       char *cwd;
-       int dircache_enabled;
        int auto_reconnect;
+       int dircache_enabled;
        struct nfsdir *dircache;
        uint16_t	mask;
-
        int auto_traverse_mounts;
        struct nested_mounts *nested_mounts;
+       int version;
+       int nfsport;
+       int mountport;
 
-        int version;
-        int nfsport;
-        int mountport;
+       /* NFSv4 specific fields */
+       verifier4 verifier;
+       char *client_name;
+       uint64_t clientid;
+       verifier4 setclientid_confirm;
+       uint32_t open_counter;
+       int has_lock_owner;
+#ifdef HAVE_MULTITHREADING
+       int multithreading_enabled;
+       libnfs_thread_t service_thread;
+       libnfs_mutex_t nfs_mutex;
+       libnfs_mutex_t nfs4_open_mutex;
+       struct nfs_thread_context *thread_ctx;
+#endif /* HAVE_MULTITHREADING */
+};
+        
+struct nfs_context {
+       struct rpc_context *rpc;
+       struct nfs_context_internal *nfsi;
+       char *error_string;
 
-        /* NFSv4 specific fields */
-        verifier4 verifier;
-        char *client_name;
-        uint64_t clientid;
-        verifier4 setclientid_confirm;
-        uint32_t seqid;
-        int has_lock_owner;
+#ifdef HAVE_MULTITHREADING
+       struct nfs_context *master_ctx;
+#endif /* HAVE_MULTITHREADING */
 };
 
+#ifdef HAVE_MULTITHREADING
+struct nfs_thread_context {
+        struct nfs_thread_context *next;
+        nfs_tid_t tid;
+        struct nfs_context nfs;
+};
+#endif /* HAVE_MULTITHREADING */
+        
 typedef int (*continue_func)(struct nfs_context *nfs, struct nfs_attr *attr,
 			     struct nfs_cb_data *data);
 
@@ -384,7 +422,9 @@ struct nfsfh {
 
         /* NFSv4 */
         struct stateid stateid;
+        uint32_t lock_owner;
         /* locking */
+        uint32_t open_seqid;
         uint32_t lock_seqid;
         struct stateid lock_stateid;
 };
@@ -560,6 +600,8 @@ int nfs4_write_async(struct nfs_context *nfs, struct nfsfh *nfsfh,
                      uint64_t count, const void *buf, nfs_cb cb,
                      void *private_data);
 
+int rpc_write_to_socket(struct rpc_context *rpc);
+        
 #ifdef __cplusplus
 }
 #endif
