@@ -23,6 +23,14 @@
 #include "aros_compat.h"
 #endif
 
+#ifdef PS2_EE
+#include "ps2_compat.h"
+#endif
+
+#ifdef PS3_PPU
+#include "ps3_compat.h"
+#endif
+
 #ifdef WIN32
 #include <win32/win32_compat.h>
 #endif
@@ -152,7 +160,13 @@ struct rpc_pdu *rpc_allocate_pdu2(struct rpc_context *rpc, int program, int vers
 		return NULL;
 	}
 	memset(pdu, 0, pdu_size);
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_lock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 	pdu->xid                = rpc->xid++;
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 	pdu->cb                 = cb;
 	pdu->private_data       = private_data;
 	pdu->zdr_decode_fn      = zdr_decode_fn;
@@ -214,12 +228,19 @@ void rpc_free_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu)
 
 void rpc_set_next_xid(struct rpc_context *rpc, uint32_t xid)
 {
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_lock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 	rpc->xid = xid;
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 }
 
 int rpc_queue_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu)
 {
-	int size, recordmarker;
+	int size;
+        int32_t recordmarker;
 
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
@@ -255,8 +276,14 @@ int rpc_queue_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu)
 		}
 
 		hash = rpc_hash_xid(pdu->xid);
+#ifdef HAVE_MULTITHREADING
+		nfs_mt_mutex_lock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 		rpc_enqueue(&rpc->waitpdu[hash], pdu);
 		rpc->waitpdu_len++;
+#ifdef HAVE_MULTITHREADING
+		nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 		return 0;
 	}
 
@@ -266,8 +293,19 @@ int rpc_queue_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu)
 	zdr_int(&pdu->zdr, &recordmarker);
 
 	pdu->outdata.size = size;
-	rpc_enqueue(&rpc->outqueue, pdu);
-
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_lock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
+        rpc_enqueue(&rpc->outqueue, pdu);
+#ifdef HAVE_MULTITHREADING
+        if (rpc->outqueue.head == pdu) {
+                nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+                rpc_write_to_socket(rpc);
+        } else {
+                nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+        }
+#endif /* HAVE_MULTITHREADING */
+        
 	return 0;
 }
 
@@ -440,7 +478,8 @@ static int rpc_process_call(struct rpc_context *rpc, ZDR *zdr)
         if (endpoint == NULL) {
 		rpc_set_error(rpc, "No endpoint found for CALL "
                               "program:0x%08x version:%d\n",
-                              call.body.cbody.prog, call.body.cbody.vers);
+                              (int)call.body.cbody.prog,
+                              (int)call.body.cbody.vers);
                 if (!found_program) {
                         return rpc_send_error_reply(rpc, &call, PROG_UNAVAIL,
                                                     0, 0);
@@ -470,7 +509,8 @@ int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size)
 	struct rpc_pdu *pdu, *prev_pdu;
 	struct rpc_queue *q;
 	ZDR zdr;
-	int pos, recordmarker = 0;
+	int pos;
+        int32_t recordmarker = 0;
 	unsigned int hash;
 	uint32_t xid;
 	char *reasbuf = NULL;
@@ -505,6 +545,11 @@ int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size)
 		zdr_destroy(&zdr);
 		for (fragment = rpc->fragments; fragment; fragment = fragment->next) {
 			total += fragment->size;
+                        if (total < fragment->size) {
+                                rpc_set_error(rpc, "Fragments too large");
+                                rpc_free_all_fragments(rpc);
+                                return -1;
+                        }
 		}
 
 		reasbuf = malloc(total);
@@ -535,7 +580,7 @@ int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size)
         }
 
 	pos = zdr_getpos(&zdr);
-	if (zdr_int(&zdr, (int *)&xid) == 0) {
+	if (zdr_u_int(&zdr, &xid) == 0) {
 		rpc_set_error(rpc, "zdr_int reading xid failed");
 		zdr_destroy(&zdr);
 		if (reasbuf != NULL) {
@@ -547,6 +592,9 @@ int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size)
 
 	/* Look up the transaction in a hash table of our requests */
 	hash = rpc_hash_xid(xid);
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_lock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 	q = &rpc->waitpdu[hash];
 
 	/* Follow the hash chain.  Linear traverse singly-linked list,
@@ -567,6 +615,9 @@ int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size)
 				prev_pdu->next = pdu->next;
 			rpc->waitpdu_len--;
 		}
+#ifdef HAVE_MULTITHREADING
+                nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 		if (rpc_process_reply(rpc, pdu, &zdr) != 0) {
 			rpc_set_error(rpc, "rpc_procdess_reply failed");
 		}
@@ -579,6 +630,9 @@ int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size)
 		}
 		return 0;
 	}
+#ifdef HAVE_MULTITHREADING
+        nfs_mt_mutex_unlock(&rpc->rpc_mutex);
+#endif /* HAVE_MULTITHREADING */
 
 	zdr_destroy(&zdr);
 	if (reasbuf != NULL) {
